@@ -39,7 +39,8 @@ int repeat = 1;
 
 void usage(void)
 {
-	printf("memhog [-rNUM] size[kmg] [policy [nodeset]]\n");
+	printf("memhog [-fFILE] [-rNUM] size[kmg] [policy [nodeset]]\n");
+	printf("-f mmap is backed by FILE\n");
 	printf("-rNUM repeat memset NUM times\n");
 	printf("-H disable transparent hugepages\n");
 	print_policies();
@@ -100,7 +101,6 @@ int main(int ac, char **av)
 	length = memsize(av[1]);
 	if (av[2] && numa_available() < 0) {
 		printf("Kernel doesn't support NUMA policy\n");
-		exit(1);
 	} else
 		loose = 1;
 	policy = parse_policy(av[2], av[3]);
